package fos

import scala.util.parsing.combinator.syntactical.StandardTokenParsers
import scala.util.parsing.input._
import fos.SimplyTypedExtended._
import fos.Term._
import fos.Type._

/** This object implements a parser and runner for Project 2.
 */
object Main extends StandardTokenParsers {
  lexical.delimiters ++= List("(", ")", "\\", ".", ":", "=", "->", "{", "}", ",", "*", "+",
                              "=>", "|")
  lexical.reserved   ++= List("Bool", "Nat", "true", "false", "if", "then", "else", "succ",
                              "pred", "iszero", "let", "in", "fst", "snd", "fix", "letrec",
                              "case", "of", "inl", "inr", "as")


  /** t ::=          "true"
   *               | "false"
   *               | number
   *               | "succ" t
   *               | "pred" t
   *               | "iszero" t
   *               | "if" t "then" t "else" t
   *               | ident
   *               | "\" ident ":" T "." t
   *               | t t
   *               | "(" t ")"
   *               | "let" ident ":" T "=" t "in" t
   *               | "{" t "," t "}"
   *               | "fst" t
   *               | "snd" t
   *               | "inl" t "as" T
   *               | "inr" t "as" T
   *               | "case" t "of" "inl" ident "=>" t "|" "inr" ident "=>" t
   *               | "fix" t
   *               | "letrec" ident ":" T "=" t "in" t
   */

  def term: Parser[Term] =
    positioned(
      simpleTerm ~ rep(simpleTerm)    ^^ { case t ~ ts => (t :: ts).reduceLeft[Term](App.apply) }
    | failure("illegal start of term"))

  def simpleTerm: Parser[Term] =
    positioned(
      "true"          ^^^ True
    | "false"         ^^^ False
    | numericLit      ^^ { case chars => lit2Num(chars.toInt) }
    | "succ" ~ term   ^? { case "succ" ~ t => Succ(t) }
    | "pred" ~ term   ^? { case "pred" ~ t => Pred(t) }
    | "iszero" ~ term ^? { case "iszero" ~ t => IsZero(t) }
    | "if" ~ term ~ "then" ~ term ~ "else" ~ term ^? {
        case "if" ~ t1 ~ "then" ~ t2 ~ "else" ~ t3 =>
          println(s"parsed if with: $t1, $t2, $t3")
          If(t1, t2, t3)
      }
    | ident ^^ { case id => Var(id) }
    | "\\" ~ ident ~ ":" ~ typ ~ "." ~ term ^? { case "\\" ~ x ~ ":" ~ tp ~ "." ~ t => Abs(x, tp, t) }
    | "(" ~> term <~ ")"  ^^ { case t => t }
    | "let" ~ ident ~ ":" ~ typ ~ "=" ~ term ~ "in" ~ term ^? {
        case "let" ~ id ~ ":" ~ tp ~ "=" ~ t1 ~ "in" ~ t2 => App(Abs(id, tp, t2), t1)
      }
    | "{" ~ term ~ "," ~ term ~ "}" ^? { case "{" ~ t1 ~ "," ~ t2 ~ "}" => TermPair(t1, t2) }
    | "fst" ~ term                  ^? { case "fst" ~ t => First(t) }
    | "snd" ~ term                  ^? { case "snd" ~ t => Second(t) }
    | "inl" ~ term ~ "as" ~ typ    ^? { case "inl" ~ t ~ "as" ~ tp => Inl(t, tp) }
    | "inr" ~ term ~ "as" ~ typ    ^? { case "inr" ~ t ~ "as" ~ tp => Inr(t, tp) }
    | "case" ~ term ~ "of" ~ "inl" ~ ident ~ "=>" ~ term ~ "|" ~ "inr" ~ ident ~ "=>" ~ term ^? {
      case "case" ~ t1 ~ "of" ~ "inl" ~ id1 ~ "=>" ~ t2 ~ "|" ~ "inr" ~ id2 ~ "=>" ~ t3 =>
        Case(t1, id1, t2, id2, t3)
      }
    | "fix" ~ term ^? { case "fix" ~ t => Fix(t) }
    | "letrec" ~ ident ~ ":" ~ typ ~ "=" ~ term ~ "in" ~ term ^? {
        case "letrec" ~ id ~ ":" ~ tp ~ "=" ~ t1 ~ "in" ~ t2 =>
          App(Abs(id, tp, t2), Fix(Abs(id, tp, t1)))
      }
    | failure("illegal start of simple term"))

  /** Type       ::= SimpleType [ "->" Type ]
   */
  def typ: Parser[Type] =
    positioned(
      simpleType ~ opt("->" ~> typ)   ^^ {
        case t1 ~ Some(t2) => TypeFun(t1, t2)
        case t1 ~ None => t1
      }
    | failure("illegal start of type"))

  /** SimpleType ::= BaseType [ ("*" SimpleType) | ("+" SimpleType) ]
   */
  def simpleType: Parser[Type] =
    positioned(
      baseType ~ opt(("*" ~ simpleType) | ("+" ~ simpleType)) ^? {
        case bt ~ Some("*" ~ st) => TypePair(bt, st)
        case bt ~ Some("+" ~ st) => TypeSum(bt, st)
        case bt ~ None => bt
      }
    | failure("illegal start of simple type"))

  /** BaseType ::= "Bool" | "Nat" | "(" Type ")"
   */
  def baseType: Parser[Type] =
    positioned(
      "Bool" ^^^ TypeBool
    | "Nat"  ^^^ TypeNat
    | "(" ~> typ <~ ")" ^^ { case t => t })

  def lit2Num(n: Int): Term =
    if (n == 0) Zero else Succ(lit2Num(n - 1))


  def main(args: Array[String]): Unit =
    val stdin = new java.io.BufferedReader(new java.io.InputStreamReader(System.in))
    val tokens = new lexical.Scanner(stdin.readLine())
    phrase(term)(tokens) match
      case Success(trees, _) =>
        try
          println("typed: " + typeof(Nil, trees))
          for (t <- path(trees, reduce))
            println(t)
        catch
          case tperror: Exception => println(tperror.toString)
      case e =>
        println(e)

}
