package fos

import scala.util.parsing.input.Positional

/** Abstract Syntax Trees for terms. */
enum Term extends Positional:
  case True
  case False
  case Zero
  case Succ(t: Term)
  case Pred(t: Term)
  case IsZero(t: Term)
  case If(cond: Term, t1: Term, t2: Term)
  case Var(name: String)
  case Abs(v: String, tp: Type, t: Term)
  case App(t1: Term, t2: Term)
  case TermPair(t1: Term, t2: Term)
  case First(t: Term)
  case Second(t: Term)
  case Inl(t: Term, tpe: Type)
  case Inr(t: Term, tpe: Type)
  case Case(t: Term, x1: String, t1: Term, x2: String, t2: Term)
  case Fix(t: Term)
  
  override def toString: String = this match
    case True => "true"
    case False => "false"
    case Zero => "0"
    case Succ(t) => "succ " + t
    case Pred(t) => "pred " + t
    case IsZero(t) => "iszero " + t
    case If(cond, t1, t2) => "if " + cond + " then " + t1 + " else " + t2
    case Var(name) => name
    case Abs(v, tp, t) => "(\\" + v + ":" + tp + "." + t + ")"
    case App(t1, t2) =>
      t1.toString + (t2 match
        case App(_, _) => " (" + t2.toString + ")" // left-associative
        case _         => " " + t2.toString
      )
    case TermPair(t1, t2) => "{" + t1 + "," + t2 + "}"
    case First(t) => "fst " + t
    case Second(t) => "snd " + t
    case Inl(t, tpe) =>
      s"(inl $t as $tpe)"
    case Inr(t, tpe) =>
      s"(inr $t as $tpe)"
    case Case(t, x1, t1, x2, t2) =>
      s"(case $t of inl $x1 => $t1 | inr $x2 => $t2)"
    case Fix(t) =>
      s"(fix $t)"


/** Abstract Syntax Trees for types. */
enum Type extends Positional:

  case TypeBool
  case TypeNat
  case TypeFun(t1: Type, t2: Type)
  case TypePair(t1: Type, t2: Type)
  case TypeSum(t1: Type, t2: Type)

  override def toString: String = this match 
    case TypeBool => "Bool"
    case TypeNat => "Nat"
    case TypeFun(t1, t2) =>
      (t1 match
        case TypeFun(_, _) => "(" + t1 + ")" // right-associative
        case _             => t1.toString
      ) + "->" + t2
    case TypePair(t1, t2) => "(" + t1 + " * " + t2 + ")"
    case TypeSum(t1, t2) => s"($t1 + $t2)"
